@@ -35,24 +35,24 @@ from inform import (
     Error,
     Info,
 )
+import textwrap
 import collections.abc
 import re
 
 
 # Globals {{{1
-__version__ = "0.6.0"
-__released__ = "2020-09-26"
+__version__ = "1.1.0"
+__released__ = "2020-10-11"
 __all__ = ['load', 'loads', 'dump', 'dumps', 'NestedTextError']
 
 
 # Exception {{{1
 class NestedTextError(Error, ValueError):
     r'''
-    :func:`loads()` and :func:`dumps()` both raise
-    *NestedTextError* when they discover an error. *NestedTextError* subclasses both
-    the Python *ValueError* and the *Error* exception from *Inform*.
-    You can find more documentation on what you can do with this exception in the
-    `Inform documentation
+    The *load* and *dump* functions all raise *NestedTextError* when they
+    discover an error. *NestedTextError* subclasses both the Python *ValueError*
+    and the *Error* exception from *Inform*.  You can find more documentation on
+    what you can do with this exception in the `Inform documentation
     <https://inform.readthedocs.io/en/stable/api.html#exceptions>`_.
 
     The exception provides the following attributes:
@@ -151,10 +151,11 @@ class NestedTextError(Error, ValueError):
 
     Exceptions produced by *NestedText* contain a *template* attribute that
     contains the basic text of the message. You can change this message by
-    overriding the attribute when using *report*, *terminate*, or *render*.
-    *render* is like casting the exception to a string except that allows for
-    the passing of arguments.  For example, to convert a particular message to
-    Spanish, you could use something like the following.
+    overriding the attribute using the *template* argument when using *report*,
+    *terminate*, or *render*.  *render* is like casting the exception to a
+    string except that allows for the passing of arguments.  For example, to
+    convert a particular message to Spanish, you could use something like the
+    following.
 
     .. code-block:: python
 
@@ -174,16 +175,16 @@ class NestedTextError(Error, ValueError):
 # Converts NestedText into Python data hierarchies.
 
 # constants {{{2
-# build regex use to split dict items
-colon_space = ": "
-colon_newline = ":\n"
-splitters = [
-    r'"[^"\n]*"',  # "string"
-    r"'[^'\n]*'",  # 'string'
-    colon_space,
-    colon_newline,
-]
-splitter = re.compile("(" + "|".join(f"(?:{s})" for s in splitters) + ")")
+# regular expressions used to recognize dict items
+dict_item_regex = r"""
+    (?P<quote>["']?)       # leading quote character, optional
+    (?P<key>.*?)           # key
+    (?P=quote)             # matching quote character
+    \s*                    # optional white space
+    :                      # separator
+    (?:\ (?P<value>.*))?   # value
+"""
+dict_item_recognizer = re.compile(dict_item_regex, re.VERBOSE)
 
 
 # report {{{2
@@ -197,7 +198,7 @@ def report(message, line, *args, colno=None, **kwargs):
         if colno is not None:
             # build codicil that shows both the line and the preceding line
             if line.prev_line is not None:
-                codicil = [f'{line.lineno-1:>4} «{line.prev_line}»']
+                codicil = [f'{line.prev_line.lineno:>4} «{line.prev_line.text}»']
             else:
                 codicil = []
             codicil += [
@@ -210,7 +211,8 @@ def report(message, line, *args, colno=None, **kwargs):
             kwargs['codicil'] = f'{line.lineno:>4} «{line.text}»'
         kwargs['line'] = line.text
         kwargs['lineno'] = line.lineno
-        kwargs['prev_line'] = line.prev_line
+        if line.prev_line:
+            kwargs['prev_line'] = line.prev_line.text
     else:
         kwargs['culprit'] = culprits  # pragma: no cover
     raise NestedTextError(template=message, *args, **kwargs)
@@ -219,20 +221,32 @@ def report(message, line, *args, colno=None, **kwargs):
 # indentation_error {{{2
 def indentation_error(line, depth):
     assert line.depth != depth
-    report('invalid indentation.', line, colno=depth)
-
-
-# is_quoted {{{2
-def is_quoted(s):
-    return s[:1] in ['"', "'"] and s[:1] == s[-1:]
-
-
-# dequote {{{2
-def dequote(s):
-    s = s.strip()
-    if is_quoted(s):
-        return s[1:-1]
-    return s
+    prev_line = line.prev_line
+    if not line.prev_line and depth == 0:
+        msg = 'top-level content must start in column 1.'
+    elif (
+        prev_line and
+        prev_line.value and
+        prev_line.depth < line.depth and
+        prev_line.kind in ['list item', 'dict item']
+    ):
+        if prev_line.value.strip() == '':
+            obs = ', which in this case consists only of whitespace'
+        else:
+            obs = ''
+        msg = ' '.join([
+            'invalid indentation.',
+            'An indent may only follow a dictionary or list item that does',
+            f'not already have a value{obs}.'
+        ])
+    elif (
+        prev_line and
+        prev_line.depth > line.depth
+    ):
+        msg = 'invalid indentation, partial dedent'
+    else:
+        msg = 'invalid indentation.'
+    report(textwrap.fill(msg), line, colno=depth)
 
 
 # Lines class {{{2
@@ -259,9 +273,8 @@ class Lines:
             key = None
             value = None
 
-            # last line may not have trailing newline, which we count on
-            if line[-1:] != '\n':
-                line += '\n'
+            # all lines except possibly last have a trailing newline, remove it
+            line = line.rstrip('\n')
 
             # compute indentation
             stripped = line.lstrip()
@@ -274,32 +287,27 @@ class Lines:
                 depth = None
             elif stripped[:1] == "#":
                 kind = "comment"
-                value = line[1:-1].strip()
+                value = line[1:].strip()
                 depth = None
-            elif stripped == '-\n' or stripped.startswith('- '):
+            elif stripped == '-' or stripped.startswith('- '):
                 kind = "list item"
-                value = stripped[2:-1]
-            elif stripped == '>\n' or stripped.startswith('> '):
-                kind = "string"
-                value = line[depth+2:-1]
+                value = stripped[2:]
+            elif stripped == '>' or stripped.startswith('> '):
+                kind = "string item"
+                value = line[depth+2:]
             else:
-                components = splitter.split(line)
-                kind = "dict item"
-                if colon_space in components:
-                    split_loc = components.index(colon_space)
-                    key = dequote("".join(components[:split_loc]))
-                    value = "".join(components[split_loc + 1:])
-                elif colon_newline in components:
-                    # :\n is always second to last component and followed by ''
-                    key = dequote("".join(components[:-2]))
-                    value = "\n"
+                matches = dict_item_recognizer.fullmatch(stripped)
+                if matches:
+                    kind = "dict item"
+                    key = matches.group('key')
+                    value = matches.group('value')
+                    if value is None:
+                        value = ''
                 else:
                     kind = "unrecognized"
                     value = line
-                value = value[:-1]
 
             # bundle information about line
-            line = line[:-1]  # remove trailing newline
             the_line = self.Line(
                 text = line,
                 lineno = lineno+1,
@@ -309,7 +317,8 @@ class Lines:
                 value = value,
                 prev_line = prev_line,
             )
-            prev_line = line
+            if kind.endswith(' item'):
+                prev_line = the_line
 
             # check the indent for non-spaces
             if depth:
@@ -336,7 +345,10 @@ class Lines:
     # still_within_string() {{{3
     def still_within_string(self, depth):
         if self.next_line:
-            return self.next_line.kind == "string" and self.next_line.depth == depth
+            return (
+                self.next_line.kind == "string item" and
+                self.next_line.depth >= depth
+            )
 
     # depth_of_next() {{{3
     def depth_of_next(self):
@@ -367,7 +379,7 @@ def read_value(lines, depth, on_dup):
         return read_list(lines, depth, on_dup)
     if lines.type_of_next() == "dict item":
         return read_dict(lines, depth, on_dup)
-    if lines.type_of_next() == "string":
+    if lines.type_of_next() == "string item":
         return read_string(lines, depth)
     report('unrecognized line.', lines.get_next())
 
@@ -433,29 +445,70 @@ def read_string(lines, depth):
     while lines.still_within_string(depth):
         line = lines.get_next()
         data.append(line.value)
+        if line.depth != depth:
+            indentation_error(line, depth)
     return "\n".join(data)
 
 
 # read_all() {{{2
-def read_all(lines, source, on_dup):
+def read_all(lines, top, source, on_dup):
     if callable(on_dup):
         on_dup = dict(_callback_func=on_dup)
 
     with set_culprit(source):
         lines = Lines(lines)
 
-        if lines.type_of_next():
-            return read_value(lines, 0, on_dup)
+        if top == 'any':
+            if lines.type_of_next():
+                return read_value(lines, 0, on_dup)
+            else:
+                return None
+
+        next_is = lines.type_of_next()
+
+        if top == 'dict':
+            if next_is == "dict item":
+                return read_dict(lines, 0, on_dup)
+            elif next_is:
+                report('content must start with key.', lines.get_next())
+            else:
+                return {}
+
+        if top == 'list':
+            if next_is == "list item":
+                return read_list(lines, 0, on_dup)
+            elif next_is:
+                report('content must start with dash (-).', lines.get_next())
+            else:
+                return []
+
+        if top == 'str':
+            if next_is == "string item":
+                return read_string(lines, 0)
+            elif next_is:
+                report('content must start with greater-than sign (>).', lines.get_next())
+            else:
+                return ""
+
+        raise NotImplementedError(top)
 
 
 # loads() {{{2
-def loads(content, source=None, *, on_dup=None):
+def loads(content, top='dict', *, source=None, on_dup=None):
     r'''
     Loads *NestedText* from string.
 
     Args:
         content (str):
             String that contains encoded data.
+        top (str):
+            Top-level data type. The NestedText format allows for a dictionary,
+            a list, or a string as the top-level data container.  By specifying
+            top as 'dict', 'list', or 'str' you constrain both the type of
+            top-level container and the return value of this function. By
+            specifying 'any' you enable support for all three data types, with
+            the type of the returned value matching that of top-level container
+            in content.
         source (str or Path):
             If given, this string is attached to any error messages as the
             culprit. It is otherwise unused. Is often the name of the file that
@@ -463,9 +516,10 @@ def loads(content, source=None, *, on_dup=None):
         on_dup (str or func):
             Indicates how duplicate keys in dictionaries should be handled. By
             default they raise exceptions. Specifying 'ignore' causes them to be
-            ignored. Specifying 'replace' results in them replacing earlier
-            items. By specifying a function, the keys can be de-duplicated.
-            This call-back function returns a new key and takes four arguments:
+            ignored (first wins). Specifying 'replace' results in them replacing
+            earlier items (last wins). By specifying a function, the keys can be
+            de-duplicated.  This call-back function returns a new key and takes
+            four arguments:
 
             1. The new key (duplicates an existing key).
             2. The new value.
@@ -476,7 +530,11 @@ def loads(content, source=None, *, on_dup=None):
                retained between multiple calls to this call back function.
 
     Returns:
-        The extracted data.  If content is empty, None is returned.
+        The extracted data.  The type of the return value is specified by the
+        top argument.  If top is 'any', then the return value will match that of
+        top-level data container in the input content. If content is empty, an
+        empty data value is return of the type specified by top. If top is
+        'any', None is returned.
 
     Raises:
         NestedTextError: if there is a problem in the *NextedText* content.
@@ -496,25 +554,25 @@ def loads(content, source=None, *, on_dup=None):
             ... """
 
             >>> try:
-            ...     data = nt.loads(contents)
+            ...     data = nt.loads(contents, 'dict')
             ... except nt.NestedTextError as e:
             ...     e.terminate()
 
             >>> print(data)
             {'name': 'Kristel Templeton', 'sex': 'female', 'age': '74'}
 
-        *loads()* takes an optional second argument, *culprit*. If specified,
-        it will be prepended to any error messages. It is often used to
-        designate the source of *contents*. For example, if *contents* were
-        read from a file, *culprit* would be the file name.  Here is a typical
-        example of reading *NestedText* from a file:
+        *loads()* takes an optional argument, *source*. If specified, it is
+        added to any error messages. It is often used to designate the source
+        of *contents*. For example, if *contents* were read from a file,
+        *culprit* would be the file name.  Here is a typical example of reading
+        *NestedText* from a file:
 
         .. code-block:: python
 
             >>> filename = 'examples/duplicate-keys.nt'
             >>> try:
             ...     with open(filename, encoding='utf-8') as f:
-            ...         addresses = nt.loads(f.read(), filename)
+            ...         addresses = nt.loads(f.read(), source=filename)
             ... except nt.NestedTextError as e:
             ...     print(e.render())
             ...     print(*e.get_codicil(), sep="\n")
@@ -562,11 +620,11 @@ def loads(content, source=None, *, on_dup=None):
 
     '''
 
-    return read_all(content.splitlines(True), source, on_dup)
+    return read_all(content.splitlines(True), top, source, on_dup)
 
 
 # load() {{{2
-def load(f=None, on_dup=None):
+def load(f=None, top='dict', *, on_dup=None):
     r'''
     Loads *NestedText* from file or stream.
 
@@ -584,11 +642,14 @@ def load(f=None, on_dup=None):
             object is given, it will be read and not closed; utf-8 encoding
             should be used..  If an iterator is given, it should generate full
             lines in the same manner that iterating on a file descriptor would.
+        top:
+            See :func:`loads` description of this argument.
         on_dup:
             See :func:`loads` description of this argument.
 
     Returns:
-        The extracted data.  If content is empty, None is returned.
+        The extracted data.
+        See :func:`loads` description of the return value.
 
     Raises:
         NestedTextError: if there is a problem in the *NextedText* content.
@@ -602,13 +663,14 @@ def load(f=None, on_dup=None):
 
             >>> import nestedtext as nt
             >>> print(open('examples/groceries.nt').read())
-            - Bread
-            - Peanut butter
-            - Jam
+            groceries:
+              - Bread
+              - Peanut butter
+              - Jam
             <BLANKLINE>
 
             >>> nt.load('examples/groceries.nt')
-            ['Bread', 'Peanut butter', 'Jam']
+            {'groceries': ['Bread', 'Peanut butter', 'Jam']}
 
         Load from a `pathlib.Path`:
 
@@ -616,7 +678,7 @@ def load(f=None, on_dup=None):
 
             >>> from pathlib import Path
             >>> nt.load(Path('examples/groceries.nt'))
-            ['Bread', 'Peanut butter', 'Jam']
+            {'groceries': ['Bread', 'Peanut butter', 'Jam']}
 
         Load from an open file object:
 
@@ -625,7 +687,7 @@ def load(f=None, on_dup=None):
             >>> with open('examples/groceries.nt') as f:
             ...     nt.load(f)
             ...
-            ['Bread', 'Peanut butter', 'Jam']
+            {'groceries': ['Bread', 'Peanut butter', 'Jam']}
 
     '''
 
@@ -636,17 +698,16 @@ def load(f=None, on_dup=None):
 
     if isinstance(f, collections.abc.Iterator):
         source = getattr(f, 'name', None)
-        return read_all(f, source, on_dup)
+        return read_all(f, top, source, on_dup)
     else:
         source = str(f)
         with open(f, encoding='utf-8') as fp:
-            return read_all(fp, source, on_dup)
+            return read_all(fp, top, source, on_dup)
 
 
 # NestedText Writer {{{1
 # Converts Python data hierarchies to NestedText.
 
-
 # render_key {{{2
 def render_key(s):
     if not is_str(s):
@@ -666,13 +727,19 @@ def render_key(s):
         or ': ' in s
         or s[:1] + s[-1:] in ['""', "''"]
     ):
-        if '"' in s and "'" in s:
-            raise NestedTextError(
-                s,
-                template = """keys that require quoting must not contain both " and '.""",
-                culprit = s,
-            )
-        return repr(s)
+        if "'" in s:
+            quotes = '"', "'"
+        else:
+            quotes = "'", '"'
+
+        # try extracting key using various both quote characters
+        # if extracted key matches given key, accept
+        for quote_char in quotes:
+            key = quote_char + s + quote_char
+            matches = dict_item_recognizer.fullmatch(key + ':')
+            if matches and matches.group('key') == s:
+                return key
+        raise NestedTextError(s, template = "cannot disambiguate key.", culprit = key)
     return s
 
 
@@ -724,6 +791,8 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
             specified in *renderers* are allowed. If *default* is not specified
             then a broader collection of value types are supported, including
             *None*, *bool*, *int*, *float*, and *list*- and *dict*-like objects.
+            In this case Booleans is rendered as 'True' and 'False' and None and
+            empty lists and dictionaries are rendered as empty strings.
         level (int):
             The number of indentation levels.  When dumps is invoked recursively
             this is used to increment the level and so the indent.  Generally
@@ -737,9 +806,6 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
         NestedTextError: if there is a problem in the input data.
 
     Examples:
-        This example writes to a string, but it is common to write to a file.
-        The file name and extension are arbitrary. However, by convention a
-        '.nt' suffix is generally used for *NestedText* files.
 
         .. code-block:: python
 
@@ -759,15 +825,17 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
             sex: female
             age: 74
 
-        The *NestedText* format only supports dictionaries, lists, and strings.
-        By default, *dumps* is configured to be rather forgiving, so it will
-        render many of the base Python data types, such as *None*, *bool*,
-        *int*, *float* and list-like types such as *tuple* and *set* by
-        converting them to the types supported by the format.  This implies
-        that a round trip through *dumps* and *loads* could result in the types
-        of values being transformed. You can prevent this by passing
+        The *NestedText* format only supports dictionaries, lists, and strings
+        and all leaf values must be strings.  By default, *dumps* is configured
+        to be rather forgiving, so it will render many of the base Python data
+        types, such as *None*, *bool*, *int*, *float* and list-like types such
+        as *tuple* and *set* by converting them to the types supported by the
+        format.  This implies that a round trip through *dumps* and *loads*
+        could result in the types of values being transformed. You can restrict
+        *dumps* to only supporting the native types of *NestedText* by passing
         `default='strict'` to *dumps*.  Doing so means that values that are not
-        dictionaries, lists, or strings generate exceptions.
+        dictionaries, lists, or strings generate exceptions; as do empty
+        dictionaries and lists.
 
         .. code-block:: python
 
@@ -868,8 +936,8 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
 
     # define object type identification functions
     if default == 'strict':
-        is_a_dict = lambda obj: isinstance(obj, dict)
-        is_a_list = lambda obj: isinstance(obj, list)
+        is_a_dict = lambda obj: (obj or level == 0) and isinstance(obj, dict)
+        is_a_list = lambda obj: (obj or level == 0) and isinstance(obj, list)
         is_a_str = lambda obj: isinstance(obj, str)
         is_a_scalar = lambda obj: False
     else:
@@ -919,16 +987,17 @@ def dumps(obj, *, sort_keys=False, indent=4, renderers=None, default=None, level
             need_indented_block = True
         else:
             content = obj
-            if level == 0:
-                content = add_leader(content, '> ').strip()
     elif is_a_scalar(obj):
-        content = str(obj)
+        if obj is None:
+            content = ''
+        else:
+            content = str(obj)
     elif default and callable(default):
         content = default(obj)
     else:
         error = "unsupported type."
 
-    if need_indented_block and level != 0:
+    if need_indented_block and content and level:
         content = "\n" + add_leader(content, indent*' ')
 
     if error:
@@ -954,7 +1023,10 @@ def dump(obj, f, **kwargs):
             closed.
 
             The name used for the file is arbitrary but it is tradition to use a
-            .nt suffix.
+            .nt suffix.  If you also wish to further distinguish the file type
+            by giving the schema, it is recommended that you use two suffixes,
+            with the suffix that specifies the schema given first and .nt given
+            last. For example: flicker.sig.nt.
 
         kwargs:
             See :func:`dumps` for optional arguments.
@@ -985,21 +1057,22 @@ def dump(obj, f, **kwargs):
             ...     with open('data.nt', 'w', encoding='utf-8') as f:
             ...         nt.dump(data, f)
             ... except nt.NestedTextError as e:
-            ...     fatal(e)
+            ...     e.terminate()
             ... except OSError as e:
             ...     fatal(os_error(e))
 
-        This example writes to a file specified by file name.
+        This example writes to a file specified by file name.  In general, the
+        file name and extension are arbitrary. However, by convention a
+        '.nt' suffix is generally used for *NestedText* files.
 
         .. code-block:: python
 
             >>> try:
             ...     nt.dump(data, 'data.nt')
             ... except nt.NestedTextError as e:
-            ...     fatal(e)
+            ...     e.terminate()
             ... except OSError as e:
             ...     fatal(os_error(e))
-            >>> data = {'key': 42, 'value': 3.1415926, 'valid': True}
 
     """
     content = dumps(obj, **kwargs)
